#include <QtCore>
#include <QDomDocument>
#include <QDomElement>

// Find references to other elements, and store them in the supplied QSet
    QSet< QString > &
findReferences(QDomElement e, QSet< QString > &references)
{
    QDomElement sibling = e;

    QRegExp urlRE("url\\(#([^)]+)\\)");

    while (!sibling.isNull()) {
	QString href = sibling.attribute("xlink:href", 
		sibling.attribute("href"));
	if (href.startsWith('#')) {
	    references << href.mid(1);
	}

	if (sibling.hasAttribute("style")) {
	    if (sibling.attribute("style").contains(urlRE)) {
		references << urlRE.cap(1);
	    }
	}
	if (sibling.hasAttribute("stroke")) {
	    if (sibling.attribute("stroke").contains(urlRE)) {
		references << urlRE.cap(1);
	    }
	}
	if (sibling.hasAttribute("fill")) {
	    if (sibling.attribute("fill").contains(urlRE)) {
		references << urlRE.cap(1);
	    }
	}
	if (sibling.hasAttribute("mask")) {
	    if (sibling.attribute("mask").contains(urlRE)) {
		references << urlRE.cap(1);
	    }
	}

	if (sibling.hasChildNodes()) {
	    findReferences(
		    sibling.firstChildElement(), references);
	}
	sibling = sibling.nextSiblingElement();
    }
    return references;
}

// Remove id attributes that are not referenced by other elements
    void
removeUnreferencedIds(QDomElement e, QSet< QString > &references)
{
    QDomElement sibling = e;

    while (!sibling.isNull()) {
	if (sibling.hasChildNodes()) {
	    removeUnreferencedIds(
		    sibling.firstChildElement(), references);
	}
	QString id = sibling.attribute("id");
	if (!references.contains(id)) {
	    //if (!id.isEmpty()) qDebug() << id;
	    sibling.removeAttribute("id");
	}

	sibling = sibling.nextSiblingElement();
    }
}

    void
removeUnreferencedDefs(QDomElement e)
{
    QDomElement sibling = e.firstChildElement();

    QList< QDomElement > unreferenced;
    int count = 0;
    while (!sibling.isNull()) {
	++count;
	if (!sibling.hasAttribute("id")) {
	    unreferenced << sibling;
	}

	sibling = sibling.nextSiblingElement();
    }

    foreach(QDomElement u, unreferenced) {
	e.removeChild(u);
    }
}

    void
stripInkscapeTags(QDomElement e)
{
    QDomElement sibling = e.firstChildElement();

    QRegExp inkscapeRE("^(i|inkscape|sodipodi):.*");
    QList< QDomElement > unwanted;
    int count = 0;
    while (!sibling.isNull()) {
	++count;
	if (inkscapeRE.exactMatch(sibling.tagName())) {
	    unwanted << sibling;
	}

	sibling = sibling.nextSiblingElement();
    }

    foreach(QDomElement u, unwanted) {
	e.removeChild(u);
    }
}

// Recursively remove attributes in the inkscape/sodipodi namespace
    void 
stripInkscapeAttributes(QDomElement e)
{
    QRegExp inkscapeRE("^(i|inkscape|sodipodi):.*");
    QDomElement sibling = e;

    while (!sibling.isNull()) {
	if (sibling.hasAttributes()) {
	    QDomNamedNodeMap attrs = sibling.attributes();
	    QStringList al;
	    for (int i = 0; i < attrs.count(); ++i) {
		QString name = attrs.item(i).nodeName();
		if (inkscapeRE.exactMatch(name)) {
		    al << name;
		}
	    }
	    for (int i = 0; i < al.count(); ++i) {
		attrs.removeNamedItem(al[i]);
	    }
	    al.clear();
	}
	if (sibling.hasChildNodes()) {
	    stripInkscapeAttributes(sibling.firstChildElement());
	}
	sibling = sibling.nextSiblingElement();
    }
}


// Recursively find floating point numbers and rewrite them in a compact form
// At most precision digits are preserved after the decimal point.
//
    void 
trimDecimalNumbers(QDomElement e, int precision = -1)
{
    QRegExp numberRE("[0-9]*\\.[0-9]+");
    QDomElement sibling = e;

    while (!sibling.isNull()) {
	if (sibling.hasAttributes()) {
	    QDomNamedNodeMap attrs = sibling.attributes();
	    for (int i = 0; i < attrs.count(); ++i) {
		QDomAttr attr = attrs.item(i).toAttr();
		QString val = attr.value();
		QStringList numbers;
		int match = 0;
		while (match != -1) {
		    match = val.indexOf(numberRE, match);
		    if (match != -1) {
			numbers << numberRE.cap();
			match += numberRE.cap().length();
		    }
		}
		foreach (QString s, numbers) {
		    double d = s.toDouble();
		    val.replace(s, QString::number(d, 'g', precision));
		}
		attr.setValue(val);
	    }
	}
	if (sibling.hasChildNodes()) {
	    trimDecimalNumbers(sibling.firstChildElement());
	}
	sibling = sibling.nextSiblingElement();
    }
}

// Remove redundant styles
    void 
stripElementList(QDomDocument &d, const QString &type)
{
    QRegExp inkScapeRE("^(inkscape|sodipodi):.*");
    QRegExp strokeRE("stroke:\\s*none;");
    QRegExp uselessStrokeStylesRE("stroke-(width|line(cap|join)|miterlimit|dash(array|offset)|opacity):[^;]*;");
    QRegExp uselessStrokeTerminalStyleRE("stroke-(width|line(cap|join)|miterlimit|dash(array|offset)|opacity):.*$");

    QRegExp strokeDashArrayRE("stroke-dasharray:\\s*none;");
    QRegExp uselessStrokeDashArrayStylesRE("stroke-dashoffset:[^;]*;");

    QRegExp fillRE("fill:\\s*none;");
    QRegExp uselessFillStylesRE("fill-(opacity|rule):[^;]*;");

    QRegExp markerRE("marker:\\s*none;");
    QRegExp uselessMarkerStylesRE("marker-(start|mid|end):[^;]*;");

    QRegExp fontStlesRE("font-(size|weight|family):[^;]*");

    QRegExp visibilityRE("\\bvisibility:visible\\b");
    QRegExp danglingSemicolonRE(";$");

    // A color that can be represented in the compact #RGB form (vs #RRGGBB)
    QRegExp colorRE("#([A-Fa-f0-9])\\1([A-Fa-f0-9])\\2([A-Fa-f0-9])\\3");
    QRegExp longColorRE("#([A-Fa-f0-9]{6,6}");

    QDomNodeList l = d.elementsByTagName(type);

    for (int i = 0; i < l.count(); ++i) {
	QDomElement n = l.at(i).toElement();
	if (n.hasAttribute("style")) {
	    QString style = n.attribute("style");

	    if (style.contains("visibility:visible")) {
		style.remove("visibility:visible");
		style.remove(uselessStrokeTerminalStyleRE);
	    }
	    if (style.contains(strokeRE)) {
		style.remove(uselessStrokeStylesRE);
		style.remove(uselessStrokeTerminalStyleRE);
	    }
	    if (style.contains(strokeDashArrayRE)) {
		style.remove(uselessStrokeDashArrayStylesRE);
	    }
	    if (style.contains(fillRE)) {
		style.remove(uselessFillStylesRE);
	    }
	    if (style.contains(markerRE)) {
		style.remove(uselessMarkerStylesRE);
	    }
	    if (type != "text") {
		if (style.contains(fontStlesRE)) {
		    style.remove(fontStlesRE);
		}
	    }
	    while (style.contains(colorRE)) {
		QString original = colorRE.cap(0);
		//qDebug() << original;
		QString color = QString('#') + original[1] + original[3] + original[5];
		style.replace(original, color.toLower());
	    }
	    while (style.contains(longColorRE)) {
		QString original = longColorRE.cap(0);
		style.replace(original, original.toLower());
	    }
	    // remove duplicate ;;
	    style.replace(";;", ";");

	    // remove ; if it is the last of the line
	    style.remove(danglingSemicolonRE);

	    n.setAttribute("style", style);
	}
	if (n.hasAttribute("color")) {
	    QString color = n.attribute("color");
	    if (color.contains(colorRE)) {
		QString cColor = '#' + colorRE.cap(1) + colorRE.cap(2) + colorRE.cap(3);
		color.replace(colorRE, cColor.toLower());
		n.setAttribute("color", color);
	    }
	}
	if (n.hasAttribute("stroke")) {
	    QString color = n.attribute("stroke");
	    if (color.contains(colorRE)) {
		QString cColor = '#' + colorRE.cap(1) + colorRE.cap(2) + colorRE.cap(3);
		color.replace(colorRE, cColor.toLower());
		n.setAttribute("stroke", color);
	    }
	}
	if (n.hasAttribute("fill")) {
	    QString color = n.attribute("fill");
	    if (color.contains(colorRE)) {
		QString cColor = '#' + colorRE.cap(1) + colorRE.cap(2) + colorRE.cap(3);
		color.replace(colorRE, cColor.toLower());
		n.setAttribute("fill", color);
	    }
	}
    }
}

// Change color specifications of the form #RRGGBB with the shorter form #RGB, when possible

    int
main(int argc, char **argv)
{
    QCoreApplication a(argc, argv);

    QDomDocument d("svg");
    QSet< QString > references;

    QFile in;
    if (!in.open(stdin, QIODevice::ReadOnly)) {
	qFatal("cannot open stdin");
	return -1;
    }
    if (!d.setContent(&in)) {
	in.close();
	return -1;
    }

    QStringList args = a.arguments();

    // Parse command line

    if (args.contains("--refs")) {
	QString refsFileName = args[args.indexOf("--refs") + 1];
	QFile refsFile(refsFileName);
	if (!refsFile.open(QIODevice::ReadOnly)) {
	    qFatal("Cannot open %s", qPrintable(refsFileName));
	    return -1;
	}
	QTextStream refsStream(&refsFile);
	while (!refsStream.atEnd()) {
	    QString ref;
	    refsStream >> ref;
	    references << ref;
	}

    }

    findReferences(d.documentElement(), references);
    //qDebug() << references;

    removeUnreferencedIds(d.documentElement(), references);

    QDomNodeList defs = d.elementsByTagName("defs");
    QDomElement def = defs.at(0).toElement();
    int i = 0;
    while (i < defs.count()) {
	QDomElement def = defs.at(i).toElement();
	removeUnreferencedDefs(def);
	i++;
    }

    stripInkscapeAttributes(d.documentElement());
    stripInkscapeTags(d.documentElement());
    
    trimDecimalNumbers(d.documentElement());

    stripElementList(d, "path");

    stripElementList(d, "rect");

    stripElementList(d, "text");

    stripElementList(d, "stop");

    in.close();
    QString outContent = d.toString(0);

    // dump the DOM to stdout

    QFile out;
    out.open(stdout, QIODevice::WriteOnly);
    QTextStream outStream(&out);

    outStream << outContent;

    out.close();

    return 0;
}
